@@ -18,15 +18,18 @@
  */
 
 #include "packageloader.h"
+#include "private/packageloader_p.h"
 
 #include <QStandardPaths>
+#include <QDateTime>
 #include <QDirIterator>
 #include <QJsonDocument>
 #include <QJsonArray>
 #include <QCoreApplication>
 #include <QPointer>
-#include <QDebug>
+#include "kpackage_debug.h"
 
+#include <KCompressionDevice>
 #include <KLocalizedString>
 #include <KPluginLoader>
 #include <KPluginFactory>
@@ -36,35 +39,19 @@
 #include "private/packages_p.h"
 #include "package.h"
 #include "packagestructure.h"
+#include "private/packagejobthread_p.h"
 
 namespace KPackage
 {
 
 static PackageLoader *s_packageTrader = nullptr;
 
-class PackageLoaderPrivate
-{
-public:
-    PackageLoaderPrivate()
-        : isDefaultLoader(false),
-          packageStructurePluginDir(QStringLiteral("kpackage/packagestructure"))
-    {
-    }
-
-    static QSet<QString> knownCategories();
-    static QString parentAppConstraint(const QString &parentApp = QString());
-
-    static QSet<QString> s_customCategories;
-    QHash<QString, QPointer<PackageStructure> > structures;
-    bool isDefaultLoader;
-    QString packageStructurePluginDir;
-};
 
 QSet<QString> PackageLoaderPrivate::s_customCategories;
 
 QSet<QString> PackageLoaderPrivate::knownCategories()
 {
-    // this is to trick the tranlsation tools into making the correct
+    // this is to trick the translation tools into making the correct
     // strings for translation
     QSet<QString> categories = s_customCategories;
     categories << QStringLiteral(I18N_NOOP("Accessibility")).toLower()
@@ -112,7 +99,7 @@ PackageLoader::PackageLoader()
 
 PackageLoader::~PackageLoader()
 {
-    foreach (auto wp, d->structures) {
+    for (auto wp : qAsConst(d->structures)) {
         delete wp.data();
     }
     delete d;
@@ -124,7 +111,7 @@ void PackageLoader::setPackageLoader(PackageLoader *loader)
         s_packageTrader = loader;
     } else {
 #ifndef NDEBUG
-        // qDebug() << "Cannot set packageTrader, already set!" << s_packageTrader;
+        // qCDebug(KPACKAGE_LOG) << "Cannot set packageTrader, already set!" << s_packageTrader;
 #endif
     }
 }
@@ -169,7 +156,7 @@ Package PackageLoader::loadPackage(const QString &packageFormat, const QString &
     }
 
 #ifndef NDEBUG
-        // qDebug() << "Couldn't load Package for" << packageFormat << "! reason given: " << error;
+        // qCDebug(KPACKAGE_LOG) << "Couldn't load Package for" << packageFormat << "! reason given: " << error;
 #endif
 
     return Package();
@@ -177,12 +164,28 @@ Package PackageLoader::loadPackage(const QString &packageFormat, const QString &
 
 QList<KPluginMetaData> PackageLoader::listPackages(const QString &packageFormat, const QString &packageRoot)
 {
+    // Note: Use QDateTime::currentSecsSinceEpoch() once we can depend on Qt 5.8
+    const qint64 now = qRound64(QDateTime::currentMSecsSinceEpoch() / 1000.0);
+    bool useRuntimeCache = true;
+    if (now - d->pluginCacheAge > d->maxCacheAge && d->pluginCacheAge != 0) {
+        // cache is old and we're not within a few seconds of startup anymore
+        useRuntimeCache = false;
+        d->pluginCache.clear();
+    }
+    QString cacheKey = QString(QStringLiteral("%1.%2")).arg(packageFormat, packageRoot);
+    if (useRuntimeCache && d->pluginCache.contains(cacheKey)) {
+        return d->pluginCache.value(cacheKey);
+    }
+    if (d->pluginCacheAge == 0) {
+        d->pluginCacheAge = now;
+    }
+
     QList<KPluginMetaData> lst;
 
     //has been a root specified?
     QString actualRoot = packageRoot;
 
-    //try to take it from the ackage structure
+    //try to take it from the package structure
     if (actualRoot.isEmpty()) {
         PackageStructure *structure = d->structures.value(packageFormat).data();
         if (!structure) {
@@ -214,24 +217,22 @@ QList<KPluginMetaData> PackageLoader::listPackages(const QString &packageFormat,
     if (QDir::isAbsolutePath(actualRoot)) {
         paths = QStringList(actualRoot);
     } else {
-        foreach(const QString &path, QStandardPaths::standardLocations(QStandardPaths::GenericDataLocation)) {
+        const auto listPath = QStandardPaths::standardLocations(QStandardPaths::GenericDataLocation);
+        for (const QString &path : listPath) {
             paths += path + QLatin1Char('/') + actualRoot;
         }
     }
 
-    Q_FOREACH(auto const &plugindir, paths) {
-        const QString &_ixfile = plugindir + QStringLiteral("kpluginindex.json");
-        QFile indexFile(_ixfile);
-        //qDebug() << "indexfile: " << _ixfile << indexFile.exists();
-        if (indexFile.exists()) {
-            qDebug() << "kpluginindex: Using indexfile: " << _ixfile << indexFile.exists();
+    for (auto const &plugindir : qAsConst(paths)) {
+        const QString &_ixfile = plugindir + s_kpluginindex;
+        KCompressionDevice indexFile(_ixfile, KCompressionDevice::BZip2);
+        if (QFile::exists(_ixfile)) {
+            qCDebug(KPACKAGE_LOG) << "kpluginindex: Using indexfile: " << _ixfile;
             indexFile.open(QIODevice::ReadOnly);
             QJsonDocument jdoc = QJsonDocument::fromBinaryData(indexFile.readAll());
             indexFile.close();
 
-
             QJsonArray plugins = jdoc.array();
-
             for (QJsonArray::const_iterator iter = plugins.constBegin(); iter != plugins.constEnd(); ++iter) {
                 const QJsonObject &obj = QJsonValue(*iter).toObject();
                 const QString &pluginFileName = obj.value(QStringLiteral("FileName")).toString();
@@ -241,9 +242,8 @@ QList<KPluginMetaData> PackageLoader::listPackages(const QString &packageFormat,
                     lst << m;
                 }
             }
-
         } else {
-            qDebug() << "kpluginindex: Not cached" << plugindir;
+            qCDebug(KPACKAGE_LOG) << "kpluginindex: Not cached" << plugindir;
             // If there's no cache file, fall back to listing the directory
             const QDirIterator::IteratorFlags flags = QDirIterator::Subdirectories;
             const QStringList nameFilters = { QStringLiteral("metadata.json"), QStringLiteral("metadata.desktop") };
@@ -275,19 +275,21 @@ QList<KPluginMetaData> PackageLoader::listPackages(const QString &packageFormat,
         }
     }
 
+    if (useRuntimeCache) {
+        d->pluginCache.insert(cacheKey, lst);
+    }
     return lst;
 }
 
 QList<KPluginMetaData> PackageLoader::findPackages(const QString &packageFormat, const QString &packageRoot, std::function<bool(const KPluginMetaData &)> filter)
 {
     QList<KPluginMetaData> lst;
-
-    Q_FOREACH(auto const &plugin, listPackages(packageFormat, packageRoot)) {
+    const auto lstPlugins = listPackages(packageFormat, packageRoot);
+    for (auto const &plugin : lstPlugins) {
         if (!filter || filter(plugin)) {
             lst << plugin;
         }
     }
-
     return lst;
 }
 
@@ -311,8 +313,9 @@ KPackage::PackageStructure *PackageLoader::loadPackageStructure(const QString &p
     QStringList libraryPaths;
 
     const QString subDirectory = QStringLiteral("kpackage/packagestructure");
+    const auto lstPaths = QCoreApplication::libraryPaths();
 
-    Q_FOREACH (const QString &dir, QCoreApplication::libraryPaths()) {
+    for (const QString &dir : lstPaths) {
         QString d = dir + QDir::separator() + subDirectory;
         if (!d.endsWith(QDir::separator())) {
             d += QDir::separator();
@@ -320,20 +323,16 @@ KPackage::PackageStructure *PackageLoader::loadPackageStructure(const QString &p
         libraryPaths << d;
     }
 
-
     QString pluginFileName;
 
-    Q_FOREACH (const QString &plugindir, libraryPaths) {
-        const QString &_ixfile = plugindir + QStringLiteral("kpluginindex.json");
-        QFile indexFile(_ixfile);
-        if (indexFile.exists()) {
+    for (const QString &plugindir : qAsConst(libraryPaths)) {
+        const QString &_ixfile = plugindir + s_kpluginindex;
+        KCompressionDevice indexFile(_ixfile, KCompressionDevice::BZip2);
+        if (QFile::exists(_ixfile)) {
             indexFile.open(QIODevice::ReadOnly);
             QJsonDocument jdoc = QJsonDocument::fromBinaryData(indexFile.readAll());
             indexFile.close();
-
-
             QJsonArray plugins = jdoc.array();
-
             for (QJsonArray::const_iterator iter = plugins.constBegin(); iter != plugins.constEnd(); ++iter) {
                 const QJsonObject &obj = QJsonValue(*iter).toObject();
                 const QString &candidate = obj.value(QStringLiteral("FileName")).toString();
@@ -370,12 +369,13 @@ KPackage::PackageStructure *PackageLoader::loadPackageStructure(const QString &p
     }
 
     if (structure && !error.isEmpty()) {
-        qWarning() << i18n("Could not load installer for package of type %1. Error reported was: %2",
+        qCWarning(KPACKAGE_LOG) << i18n("Could not load installer for package of type %1. Error reported was: %2",
                             packageFormat, error);
     }
 
-    if (structure)
+    if (structure) {
         d->structures.insert(packageFormat, structure);
+    }
 
     return structure;
 }
